import pygame
import random
import context
from tasks import *
from const import *
from lib.euclid import Vector2, Point2, Point3, Vector3, Matrix4



class GameObject(pygame.sprite.Sprite):
    def __init__(self, level, **kwarg):
        super(GameObject, self).__init__()
        self.travel_speed = 2/TIMESTEP
        self.level = level
        self.ai_stack = context.ContextDriver()
        self.ai_lists = []
        self.ai_cursor = 0 
        self.move_vector = Vector3(0,0,0)
        self.position = Vector3(0,0,0)
        self.power = 1
        self.xp = 0
        self._hp = 10
        self.max_hp = 10.0
        self.stuck = False
        self.carried = []

        self.__dict__.update(kwarg)

        self.init()

        self.pixel_size = self.world_size * TILESIZE

    def update(self, time):
        if self.stuck:
            return

        this_task = self.ai_stack.current_context
        if this_task is None:
            self.ai_stack = context.ContextDriver()
            for task in reversed(self.ai_lists[0]):
                self.ai_stack.append(task, start=False)
            self.ai_stack.current_context.__enter__()
        else:
            this_task.update(time)

    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, hp):
        self._hp = hp

class Missile(GameObject):
    def init(self):
        self.image = pygame.Surface((4, 4))
        self.image.fill((255,255,255))
        self.world_size = Vector2(.1, .1)
        self.travel_speed = 15/TIMESTEP

        self.ai_lists.append([
            MissileMoveTask(sprite=self, target=self.target)
        ])

class Turret(GameObject):
    def init(self):
        self.image = pygame.Surface((24, 24))
        self.image.fill(colors[48])
        self.world_size = Vector2(.2, .5)

        self.ai_lists.append([
            ShootEnemiesTask(sprite=self, target=[Hunter, Thief]),
        ])

class Minion(GameObject):
    def init(self):
        self.world_size = Vector2(.2, .5)
        self._xp = 0

        self.ai_lists.append([
            CollectTask(sprite=self, key=self.__class__),
            DepositTask(sprite=self, key=self.__class__)
        ])

        self.create_image()

    def create_image(self):
        sx = int(12 * (self.hp/self.max_hp))
        sy = int(24 * (self.hp/self.max_hp))
        if sx < 2: sx = 2
        if sy < 2: sy = 2
        self.image = pygame.Surface((sx, sy))
        self.image.fill(colors[3])
        for i in range(self.power):
            pygame.draw.line(self.image, (24,33,20), (2, i*3+2), (sx-3, i*3+2), 2)


    @property
    def xp(self):
        return self._xp

    @xp.setter
    def xp(self, xp):
        self._xp = xp
        if xp >= 8:
            if self.power < 5:
                self.travel_speed *= 1.1
                self._xp = 0
                self.hp += 1
                self.power += 1
                self.create_image()

    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, hp):
        self._hp = hp
        if hp <= 0:
            self._hp = 0
            self.kill()
            return
        


class Thief(GameObject):
    def init(self):
        self.travel_speed = 1.5/TIMESTEP
        self.image = pygame.Surface((24, 48))
        self.image.fill(colors[ENERGY*16])
        self.world_size = Vector2(.2, .5)
        self._hp = 30.0
        self.max_hp = 30.0
        self.power = 10
    
        self.ai_lists.append([
            SearchAndTravelTask(sprite=self, tile_type=[ENERGY], key=self.__class__),
            MineTask(sprite=self, duration=1000.0, steps=0, target=ENERGY),
        ])

    def update(self, time):
        self.hp += len(self.carried)
        self.carried = []
        super(Thief, self).update(time)

    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, hp):
        self._hp = hp
        if hp <= 0:
            self.kill()
            return
        if hp > 60:
            self._hp = 60

        sx = int(24 * (hp/self.max_hp))
        sy = int(48 * (hp/self.max_hp))
        if sx < 2: sx = 2
        if sy < 2: sy = 2
        self.image = pygame.Surface((sx, sy))
        self.image.fill(colors[ENERGY*16])

class Hunter(GameObject):
    def init(self):
        self.travel_speed = 2.5/TIMESTEP
        self.image = pygame.Surface((24, 48))
        self.image.fill(colors[128])
        self.world_size = Vector2(.2, .5)
        self._hp = 10.0
        self.max_hp = 10.0

        self.ai_lists.append([
            HuntTask(sprite=self, target=[Minion]),
            ChompMinion(sprite=self),
        ])

    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, hp):
        self._hp = hp
        if hp <= 0:
            self.kill()
            return
        
        sx = int(24 * (hp/self.max_hp))
        sy = int(48 * (hp/self.max_hp))
        if sx < 2: sx = 2
        if sy < 2: sy = 2
        self.image = pygame.Surface((sx, sy))
        self.image.fill(colors[128])


class Shuttle(GameObject):
    def init(self):
        self.image = pygame.Surface((12, 24))
        self.image.fill(colors[-1])
        self.world_size = Vector2(.2, .5)
        self.ai_lists.append([ShuttleFlyTask(sprite=self)])

