import collections
import random

from euclid import Vector3
from pathfinding import aStar
from const import *
import context



class Task(context.Context):
    def update(self, time):
        pass

class ShuttleFlyTask(Task):
    def enter(self):
        self.sprite.move_vector = Vector3(0,0,.02/TIMESTEP)

    def update(self, time):
        self.sprite.position += self.sprite.move_vector * time

    def exit(self):
        self.sprite.move_vector = Vector3(0,0,0)

class MoveRandomTask(Task):
    def update(self, time):
        self.sprite.position += self.sprite.move_vector * time
        if random.random() > .90:
            s = self.sprite.travel_speed
            self.sprite.move_vector = Vector3(random.uniform(-s, s), random.uniform(-s, s), 0)

class StealStockpileTask(Task):
    pass

class MineTask(Task):
    duration = 150.0
    steps = 15
    empty = BARREN
    target = None

    def update(self, time):
        self._progress += time
        self.progress = self._progress / self.duration

        x, y, z = map(int, self.sprite.position)
        if not self._mining == self.sprite.level.get_tile((x, y)):
            self.done()

        elif self.progress >= 1.0:
            spr = self.sprite
            x, y, z = map(int, spr.position)
            t, v = spr.level.get_tile((x, y))
            if v >= self.steps - self.sprite.power:
                spr.level.set_tile((x, y), self.empty)
            else:
                spr.level.set_tile((x, y), t, v+1)
            spr.carried.append(t)
            spr.xp += 1
            self.done()

    def enter(self):
        if len(self.sprite.carried) >= 1:
            self.done()
        else:
            self._progress = 0.0
            self.progress = 0.0
            x, y, z = map(int, self.sprite.position)
            t, v = self.sprite.level.get_tile((x, y))
            if self.target is None or self.target == t:
                self._mining = (t, v)
            else:
                self.done()

    def exit(self):
        my_key, dest = getattr(self.parent, "release_key", (None, None))
        if my_key is self:
            try:
                del ExclusiveKeys[my_key]
            except KeyError:
                pass
        elif my_key is not None:
            try:
                ExclusiveKeys[my_key].remove(dest)
            except (AttributeError, KeyError, ValueError):
                pass

class IntervalTask(Task):
    interval = 100.0
    show = True

    def test(self):
        return True

    def update(self, time):
        if self.test():
            self.timer += time
            if self.timer <= self.interval:
                if self.show: self.progress = self.timer/self.interval
            else:
                self.timer = 0.0
                self.work()
        elif self.show:
            self.progress = -1

    def work(self):
        pass

    def enter(self):
        self.timer = 0.0

class ChompMinion(IntervalTask):
    interval = 1.0
    show = False

    def work(self):
        spr = self.sprite
        victim = self.parent.victim

        if victim is None:
            self.done()
            return

        victim.hp -= 1
        if victim.hp <= 0:
            self.parent.victim = None
            self.done()

class ShootEnemiesTask(IntervalTask):
    interval = 100.0
    spread = 3
    energy_cost = 1

    def test(self):
        group = self.sprite.groups()[0]
        return list(group.of_class(self.target)) and group.player.energy >= self.energy_cost

    def work(self):
        from gameobjects import Missile

        spr = self.sprite
        group = spr.groups()[0]


        shot = 0
        for e in group.of_class(self.target):
            if shot < self.spread and spr.groups()[0].player.energy >= self.energy_cost:
                shot += 1
                spr.groups()[0].player.energy -= self.energy_cost
                m = Missile(spr.level, target=self.target)
                m.position = spr.position.copy()
                d = spr.position - e.position
                d.normalize()
                m.move_vector = d * m.travel_speed
                group.add(m)

ExclusiveKeys = collections.defaultdict(set)

class SearchAndTravelTask(Task):
    exclusive = True

    def update(self, time):
        spr = self.sprite

        #n = spr.position - self.path[-1] - (.5, .5, 0)
        n = spr.position - self.path[-1]
        if abs(n) < .05:
            self.path.pop()
            if len(self.path) == 0:
                spr.move_vector = Vector3(0,0,0)
                spr.xp += 1
                self.done()
                return

        spr.move_vector = n.normalize()
        spr.position -= spr.move_vector * spr.travel_speed

    @property
    def targets(self):
        return self.tile_type

    def plan(self):
        spr = self.sprite

        if self.my_key is None:
            blacklist = []
        else:
            blacklist = ExclusiveKeys[self.my_key]
        self.dest, dist = spr.level.nearest_tile(spr.position, self.targets, blacklist)

        if self.dest is None:
            return False
        else:
            self.path = aStar(tuple(map(round, spr.position[:2])), tuple(self.dest), (20, 20))
            self.path = [ i + (.5,.5,0) for i in self.path ]
            if self.exclusive:
                dest = tuple(self.dest)
                ExclusiveKeys[self.my_key].add(dest)
                self.parent.release_key = (self.my_key, dest)
            return True

    def enter(self):
        self.dest = None
        self.path = None
        self.my_key = getattr(self, "key", None)
        if not self.plan():
            self.done()

class CollectTask(SearchAndTravelTask):
    @property
    def targets(self):
        return transport_map.keys()

    def enter(self):
        if len(self.sprite.carried) >= 1:
            self.done()
        else:
            super(CollectTask, self).enter()

    def exit(self):
        super(CollectTask, self).exit()
        if self.dest is not None:
            self.parent.queue(MineTask(sprite=self.sprite, duration=200.0/self.sprite.power))

class DepositTask(SearchAndTravelTask):
    exclusive = False

    @property
    def targets(self):
        try:
            t = set(transport_map[i] for i in self.sprite.carried)
        except:
            return []
        return t

    def exit(self):
        super(DepositTask, self).exit()
        if self.dest is not None:
            task = DropCarriedTask(sprite=self.sprite)
            self.parent.queue(task)

class DropCarriedTask(Task):
    empty = BARREN

    force = {
        FARM: 8,
    }

    def enter(self):
        spr = self.sprite
        if len(spr.carried) > 0:
            coords = map(int, spr.position)[:2]
            tile, value = spr.level.get_tile(coords)

            depo_map = {v:k for k, v in transport_map.items()}
            try:
                target = depo_map[tile]
            except KeyError:
                self.done()
                return

            to_remove = []
            for i, carried in enumerate(self.sprite.carried):
                if carried == target:
                    to_remove.append(i)
                    if value+self.force[target] > 15:
                        spr.level.set_tile(coords, self.empty)
                        self.filled(coords, target)
                        self.done()
                        return
                    else:
                        spr.level.set_tile(coords, tile, value + self.force[target])

            for i in to_remove:
                spr.carried.pop(i)

        self.done()

    def filled(self, coords, tile):
        if tile == FARM:
            from gameobjects import Minion
            m = Minion(self.sprite.level)
            m.position = self.sprite.position.copy()
            self.sprite.groups()[0].add(m)


class MissileMoveTask(Task):
    def update(self, time):
        spr = self.sprite
        spr.position -= spr.move_vector * spr.travel_speed
        if abs(spr.position - (10,10,0)) > 15:
            spr.kill()

        spr = self.sprite
        groups = spr.groups()
        if groups:
            enemies = groups[0].of_class(self.target)
            
            for e in enemies:
                if abs(spr.position - e.position) < .35:
                    e.hp -= 1
                    spr.kill()
                    break

class HuntTask(Task):
    def update(self, time):
        spr = self.sprite
        victim = self.parent.victim

        if victim is None:
            group = spr.groups()[0]
            victims = list(group.of_class(self.target))
            if victims:
                victim = random.choice(victims)
                self.parent.victim = victim
            else:
                self.done()
                return

        spr.move_vector = spr.position - victim.position
        spr.move_vector.normalize()
        spr.position -= spr.move_vector * spr.travel_speed

        if abs(spr.position - victim.position) < .25:
            victim.stuck = 1
            victim.position = spr.position + (.05, .05, .25)
            self.done()

    def enter(self):
        self.parent.victim = None
